#!/bin/sh

command='rsync -avc --exclude=".*" /home/rusconi/devel/libmassgui/src ./libmassgui/ $@'
printf "Running %s\n\n" "${command}"

command='rsync -avc --exclude=".*" /home/rusconi/devel/libmassgui/includes/libmassgui ./libmassgui/includes/ $@'
printf "Running %s\n\n" "${command}"
eval $command

