#!/bin/sh

command='rsync -avc --exclude=".*" /home/rusconi/devel/libmass/src ./libmass/ $@'
printf "Running %s\n\n" "${command}"

command='rsync -avc --exclude=".*" /home/rusconi/devel/libmass/includes/libmass ./libmass/includes/ $@'
printf "Running %s\n\n" "${command}"

eval $command

